from typing import Union, Optional, Dict, Any

from pydantic import Field, root_validator

from .example import Example
from .general import Reference
from .object_base import ObjectExtended
from .schemas import Schema
from .media import MediaType


class ParameterBase(ObjectExtended):
    """
    A `Parameter Object`_ defines a single operation parameter.

    .. _Parameter Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#external-documentation-object
    """

    description: Optional[str] = Field(default=None)
    required: Optional[bool] = Field(default=None)
    deprecated: Optional[bool] = Field(default=None)
    allowEmptyValue: Optional[bool] = Field(default=None)

    style: Optional[str] = Field(default=None)
    explode: Optional[bool] = Field(default=None)
    allowReserved: Optional[bool] = Field(default=None)
    schema_: Optional[Union[Schema, Reference]] = Field(default=None, alias="schema")
    example: Optional[Any] = Field(default=None)
    examples: Optional[Dict[str, Union["Example", Reference]]] = Field(default_factory=dict)

    content: Optional[Dict[str, "MediaType"]]

    @root_validator
    def validate_ParameterBase(cls, values):
        #        if values["in_"] ==
        #        if self.in_ == "path" and self.required is not True:
        #            err_msg = 'Parameter {} must be required since it is in the path'
        #            raise SpecError(err_msg.format(self.get_path()), path=self._path)
        return values


class Parameter(ParameterBase):
    name: str = Field(required=True)
    in_: str = Field(required=True, alias="in")  # TODO must be one of ["query","header","path","cookie"]


class Header(ParameterBase):
    """

    .. _HeaderObject: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#header-object
    """



from typing import Optional, Dict

from pydantic import Field, root_validator

from .object_base import ObjectExtended


class OAuthFlow(ObjectExtended):
    """
    Configuration details for a supported OAuth Flow

    .. here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#oauth-flow-object
    """

    authorizationUrl: Optional[str] = Field(default=None)
    tokenUrl: Optional[str] = Field(default=None)
    refreshUrl: Optional[str] = Field(default=None)
    scopes: Dict[str, str] = Field(default_factory=dict)


class OAuthFlows(ObjectExtended):
    """
    Allows configuration of the supported OAuth Flows.

    .. here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#oauth-flows-object
    """

    implicit: Optional[OAuthFlow] = Field(default=None)
    password: Optional[OAuthFlow] = Field(default=None)
    clientCredentials: Optional[OAuthFlow] = Field(default=None)
    authorizationCode: Optional[OAuthFlow] = Field(default=None)


class SecurityScheme(ObjectExtended):
    """
    A `Security Scheme`_ defines a security scheme that can be used by the operations.

    .. _Security Scheme: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#security-scheme-object
    """

    type: str = Field(...)
    description: Optional[str] = Field(default=None)
    name: Optional[str] = Field(default=None)
    in_: Optional[str] = Field(default=None, alias="in")
    scheme_: Optional[str] = Field(default=None, alias="scheme")
    bearerFormat: Optional[str] = Field(default=None)
    flows: Optional[OAuthFlows] = Field(default=None)
    openIdConnectUrl: Optional[str] = Field(default=None)

    @root_validator
    def validate_SecurityScheme(cls, values):
        t = values.get("type", None)
        keys = set(map(lambda x: x[0], filter(lambda x: x[1] is not None, values.items())))
        keys -= frozenset(["type", "description"])
        if t == "apikey":
            assert keys == set(["in_", "name"])
        if t == "http":
            assert keys - frozenset(["scheme_", "bearerFormat"]) == set([])
        if t == "oauth2":
            assert keys == frozenset(["flows"])
        if t == "openIdConnect":
            assert keys - frozenset(["openIdConnectUrl"]) == set([])
        return values



from typing import Union, Optional, Dict

from pydantic import Field

from .example import Example
from .object_base import ObjectExtended
from .paths import Reference, RequestBody, Link, Parameter, Response, Callback, Header, PathItem
from .schemas import Schema
from .security import SecurityScheme


class Components(ObjectExtended):
    """
    A `Components Object`_ holds a reusable set of different aspects of the OAS
    spec.

    .. _Components Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#components-object
    """

    schemas: Optional[Dict[str, Union[Schema, Reference]]] = Field(default_factory=dict)
    responses: Optional[Dict[str, Union[Response, Reference]]] = Field(default_factory=dict)
    parameters: Optional[Dict[str, Union[Parameter, Reference]]] = Field(default_factory=dict)
    examples: Optional[Dict[str, Union[Example, Reference]]] = Field(default_factory=dict)
    requestBodies: Optional[Dict[str, Union[RequestBody, Reference]]] = Field(default_factory=dict)
    headers: Optional[Dict[str, Union[Header, Reference]]] = Field(default_factory=dict)
    securitySchemes: Optional[Dict[str, Union[SecurityScheme, Reference]]] = Field(default_factory=dict)
    links: Optional[Dict[str, Union[Link, Reference]]] = Field(default_factory=dict)
    callbacks: Optional[Dict[str, Union[Callback, Reference]]] = Field(default_factory=dict)


#    pathItems: Optional[Dict[str, Union[PathItem, Reference]]] = Field(default_factory=dict) #v3.1

Components.update_forward_refs()



import json
from typing import List

import httpx
import yarl

from .paths import SecurityRequirement


class RequestParameter:
    def __init__(self, url: yarl.URL):
        self.url = str(url)
        self.auth = None
        self.cookies = {}
        self.path = {}
        self.params = {}
        self.content = None
        self.headers = {}


class Request:
    """
    This class is returned by instances of the OpenAPI class when members
    formatted like call_operationId are accessed, and a valid Operation is
    found, and allows calling the operation directly from the OpenAPI object
    with the configured values included.  This class is not intended to be used
    directly.
    """

    def __init__(self, api: "OpenAPI", method: str, path: str, operation: "Operation.request"):
        self.api = api
        self.spec = api._spec
        self.method = method
        self.path = path
        self.operation = operation
        #        self.session:Union[httpx.Client,httpx.AsyncClient] =
        self.req: RequestParameter = RequestParameter(self.path)

    def __call__(self, *args, **kwargs):
        return self.request(*args, **kwargs)

    @property
    def security(self):
        return self.api._security

    def args(self, content_type: str = "application/json"):
        op = self.operation
        parameters = op.parameters + self.spec.paths[self.path].parameters

        schema = op.requestBody.content[content_type].schema_
        #        if isinstance(schema, Reference):
        #            schema = schema._target
        return {"parameters": parameters, "data": schema}

    def return_value(self, http_status=200, content_type="application/json"):
        return self.operation.responses[str(http_status)].content[content_type].schema_

    def _prepare_security(self):
        if self.security and self.operation.security:
            for scheme, value in self.security.items():
                for r in filter(lambda x: x.name == scheme, self.operation.security):
                    self._prepare_secschemes(r, value)
                    break
                else:
                    continue
                break
            else:
                raise ValueError(
                    f"No security requirement satisfied (accepts {', '.join(self.operation.security.keys())})"
                )

    def _prepare_secschemes(self, security_requirement: SecurityRequirement, value: List[str]):
        ss = self.spec.components.securitySchemes[security_requirement.name]

        if ss.type == "http" and ss.scheme_ == "basic":
            self.req.auth = value

        if ss.type == "http" and ss.scheme_ == "digest":
            self.req.auth = httpx.DigestAuth(*value)

        if ss.type == "http" and ss.scheme_ == "bearer":
            header = ss.bearerFormat or "Bearer {}"
            self.req.headers["Authorization"] = header.format(value)

        if ss.type == "mutualTLS":
            # TLS Client certificates (mutualTLS)
            self.req.cert = value

        if ss.type == "apiKey":
            if ss.in_ == "query":
                # apiKey in query parameter
                self.req.params[ss.name] = value

            if ss.in_ == "header":
                # apiKey in query header data
                self.req.headers[ss.name] = value

            if ss.in_ == "cookie":
                self.req.cookies = {ss.name: value}

    def _prepare_parameters(self, parameters):
        # Parameters
        path_parameters = {}
        accepted_parameters = {}
        p = self.operation.parameters + self.spec.paths[self.path].parameters

        for _ in list(p):
            # TODO - make this work with $refs - can operations be $refs?
            accepted_parameters.update({_.name: _})

        for name, spec in accepted_parameters.items():
            if parameters is None or name not in parameters:
                if spec.required:
                    raise ValueError(f"Required parameter {name} not provided")
                continue

            value = parameters[name]

            if spec.in_ == "path":
                # The string method `format` is incapable of partial updates,
                # as such we need to collect all the path parameters before
                # applying them to the format string.
                path_parameters[name] = value

            if spec.in_ == "query":
                self.req.params[name] = value

            if spec.in_ == "header":
                self.req.headers[name] = value

            if spec.in_ == "cookie":
                self.req.cookies[name] = value

        self.req.url = self.req.url.format(**path_parameters)

    def _prepare_body(self, data):
        if not self.operation.requestBody:
            return

        if data is None and self.operation.requestBody.required:
            raise ValueError("Request Body is required but none was provided.")

        if "application/json" in self.operation.requestBody.content:
            if not isinstance(data, (dict, list)):
                raise TypeError(data)
            body = json.dumps(data)
            self.req.content = body.encode()
            self.req.headers["Content-Type"] = "application/json"
        else:
            raise NotImplementedError()

    def _prepare(self, data, parameters):
        self._prepare_security()
        self._prepare_parameters(parameters)
        self._prepare_body(data)

        req = httpx.Request(
            self.method,
            str(self.api.url / self.req.url[1:]),
            headers=self.req.headers,
            cookies=self.req.cookies,
            params=self.req.params,
            content=self.req.content,
        )
        return req

    def _process(self, result):
        # spec enforces these are strings
        status_code = str(result.status_code)

        # find the response model in spec we received
        expected_response = None
        if status_code in self.operation.responses:
            expected_response = self.operation.responses[status_code]
        elif "default" in self.operation.responses:
            expected_response = self.operation.responses["default"]

        if expected_response is None:
            # TODO - custom exception class that has the response object in it
            raise ValueError(
                f"""Unexpected response {result.status_code} from {self.operation.operationId} (expected one of { ",".join(self.operation.responses.keys()) }), no default is defined"""
            )

        # defined as "no content"
        if len(expected_response.content) == 0:
            return None

        content_type = result.headers.get("Content-Type", None)
        if content_type:
            expected_media = expected_response.content.get(content_type, None)
            if expected_media is None and "/" in content_type:
                # accept media type ranges in the spec. the most specific matching
                # type should always be chosen, but if we do not have a match here
                # a generic range should be accepted if one if provided
                # https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#response-object

                generic_type = content_type.split("/")[0] + "/*"
                expected_media = expected_response.content.get(generic_type, None)
        else:
            expected_media = None

        if expected_media is None:
            raise RuntimeError(
                f"Unexpected Content-Type {content_type} returned for operation {self.operation.operationId} \
                         (expected one of {','.join(expected_response.content.keys())})"
            )

        if content_type.lower() == "application/json":
            return expected_media.schema_.model(result.json())
        else:
            raise NotImplementedError()

    def request(self, data=None, parameters=None):
        """
        Sends an HTTP request as described by this Path

        :param data: The request body to send.
        :type data: any, should match content/type
        :param parameters: The parameters used to create the path
        :type parameters: dict{str: str}
        """

        req = self._prepare(data, parameters)

        result = self.api._session_factory(auth=self.req.auth).send(req)
        return self._process(result)


class AsyncRequest(Request):
    async def __call__(self, *args, **kwargs):
        return await self.request(*args, **kwargs)

    async def request(self, data=None, parameters=None):
        req = self._prepare(data, parameters)

        async with self.api._session_factory(auth=self.req.auth) as client:
            result = await client.send(req)

        return self._process(result)



from .openapi import OpenAPI
from .loader import FileSystemLoader
from .errors import SpecError, ReferenceResolutionError


__version__ = "0.1.1"
__all__ = ["__version__", "OpenAPI", "FileSystemLoader", "SpecError", "ReferenceResolutionError"]



import abc
import json
from pathlib import Path

import yaml


class Loader(abc.ABC):
    @abc.abstractmethod
    def load(self, name: str):
        raise NotImplementedError("load")

    @classmethod
    def decode(cls, data, codec):
        if codec is not None:
            codecs = [codec]
        else:
            codecs = ["ascii", "utf-8"]
        for c in codecs:
            try:
                data = data.decode(c)
                break
            except UnicodeError:
                continue
        else:
            raise ValueError("encoding")
        return data

    @classmethod
    def dict(cls, file, data):
        if file.suffix == ".yaml":
            data = yaml.safe_load(data)
        elif file.suffix == ".json":
            data = json.loads(data)
        else:
            raise ValueError(file.name)
        return data


class FileSystemLoader(Loader):
    def __init__(self, base: Path):
        assert isinstance(base, Path)
        self.base = base

    def load(self, file: str, codec=None):
        path = self.base / file
        assert path.is_relative_to(self.base)
        data = path.open("rb").read()
        data = self.decode(data, codec)
        data = self.dict(path, data)
        return data



import urllib.parse
from typing import Optional

from pydantic import Field, Extra
from yarl import URL

from .object_base import ObjectExtended, ObjectBase


class ExternalDocumentation(ObjectExtended):
    """
    An `External Documentation Object`_ references external resources for extended
    documentation.

    .. _External Documentation Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#external-documentation-object
    """

    url: str = Field(...)
    description: Optional[str] = Field(default=None)


class JSONPointer:
    @staticmethod
    def decode(part):
        """

        https://swagger.io/docs/specification/using-ref/
        :param part:
        """
        part = urllib.parse.unquote(part)
        part = part.replace("~1", "/")
        return part.replace("~0", "~")


class JSONReference:
    @staticmethod
    def split(url):
        u = URL(url)
        return str(u.with_fragment("")), u.raw_fragment


class Reference(ObjectBase):
    """
    A `Reference Object`_ designates a reference to another node in the specification.

    .. _Reference Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#reference-object
    """

    ref: str = Field(alias="$ref")

    _target: object = None

    class Config:
        """This object cannot be extended with additional properties and any properties added SHALL be ignored."""

        extra = Extra.ignore

    def __getattr__(self, item):
        if item != "_target":
            return getattr(self._target, item)
        else:
            return getattr(self, item)

    def __setattr__(self, item, value):
        if item != "_target":
            setattr(self._target, item, value)
        else:
            super().__setattr__(item, value)



from typing import Union, Optional

from pydantic import Field

from .general import Reference
from .object_base import ObjectExtended


class Example(ObjectExtended):
    """
    A `Example Object`_ holds a reusable set of different aspects of the OAS
    spec.

    .. _Example Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#example-object
    """

    summary: Optional[str] = Field(default=None)
    description: Optional[str] = Field(default=None)
    value: Optional[Union[Reference, dict, str]] = Field(default=None)  # 'any' type
    externalValue: Optional[str] = Field(default=None)



from pydantic import Field

from .general import ObjectExtended


class XML(ObjectExtended):
    """

    .. XML Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#xml-object
    """

    name: str = Field(default=None)
    namespace: str = Field(default=None)
    prefix: str = Field(default=None)
    attribute: bool = Field(default=False)
    wrapped: bool = Field(default=False)



import datetime
import pathlib
from typing import Any, List, Optional, Dict, Union, Callable

import yaml
from pydantic import Field
import httpx
import yarl

from .components import Components
from .errors import ReferenceResolutionError, SpecError
from .general import Reference, JSONPointer, JSONReference
from .info import Info
from .object_base import ObjectExtended, ObjectBase
from .paths import PathItem, SecurityRequirement, _validate_parameters, Operation
from .servers import Server
from .schemas import Schema, Discriminator
from .tag import Tag
from .request import Request, AsyncRequest
from .loader import Loader

HTTP_METHODS = frozenset(["get", "delete", "head", "post", "put", "patch", "trace"])


class OpenAPI:
    @property
    def paths(self):
        return self._spec.paths

    @property
    def components(self):
        return self._spec.components

    @property
    def info(self):
        return self._spec.info

    @property
    def openapi(self):
        return self._spec.openapi

    @property
    def servers(self):
        return self._spec.servers

    @classmethod
    def load_sync(cls, url, session_factory: Callable[[], httpx.Client] = httpx.Client, loader=None):
        resp = session_factory().get(url)
        return cls.loads(url, resp.text, session_factory, loader)

    @classmethod
    async def load_async(cls, url, session_factory: Callable[[], httpx.AsyncClient] = httpx.AsyncClient, loader=None):
        async with session_factory() as client:
            resp = await client.get(url)
        return cls.loads(url, resp.text, session_factory, loader)

    @classmethod
    def loads(cls, url, data, session_factory: Callable[[], httpx.AsyncClient] = httpx.AsyncClient, loader=None):
        data = Loader.dict(pathlib.Path(url), data)
        return cls(url, data, session_factory, loader)

    def __init__(
        self,
        url,
        raw_document,
        session_factory: Callable[[], Union[httpx.Client, httpx.AsyncClient]] = httpx.AsyncClient,
        loader=None,
    ):
        """
        Creates a new OpenAPI document from a loaded spec file.  This is
        overridden here because we need to specify the path in the parent
        class' constructor.

        :param raw_document: The raw OpenAPI file loaded into python
        :type raw_document: dct
        :param session_factory: default uses new session for each call, supply your own if required otherwise.
        :type session_factory: returns httpx.AsyncClient or http.Client
        """

        self._base_url: yarl.URL = yarl.URL(url)
        self.loader: Loader = loader
        self._session_factory = session_factory

        self._security: List[str] = None
        self._cached: Dict[str, "OpenAPISpec"] = dict()

        self._spec = OpenAPISpec.parse_obj(raw_document)
        self._spec._resolve_references(self)

        for name, schema in self.components.schemas.items():
            schema._identity = name

        operation_map = set()

        def test_operation(operation_id):
            if operation_id in operation_map:
                raise SpecError(f"Duplicate operationId {operation_id}", element=None)
            operation_map.add(operation_id)

        for path, obj in self.paths.items():
            for m in obj.__fields_set__ & HTTP_METHODS:
                op = getattr(obj, m)
                _validate_parameters(op, path)
                if op.operationId is None:
                    continue
                formatted_operation_id = op.operationId.replace(" ", "_")
                test_operation(formatted_operation_id)
                for r, response in op.responses.items():
                    if isinstance(response, Reference):
                        continue
                    for c, content in response.content.items():
                        if content.schema_ is None:
                            continue
                        if isinstance(content.schema_, Schema):
                            content.schema_._identity = f"{path}.{m}.{r}.{c}"

    @property
    def url(self):
        return self._base_url.join(yarl.URL(self._spec.servers[0].url))

    # public methods
    def authenticate(self, security_scheme, value):
        """
        Authenticates all subsequent requests with the given arguments.

        TODO - this should support more than just HTTP Auth
        """

        # authentication is optional and can be disabled
        if security_scheme is None:
            self._security = None
            return

        if security_scheme not in self._spec.components.securitySchemes:
            raise ValueError("{} does not accept security scheme {}".format(self.info.title, security_scheme))

        self._security = {security_scheme: value}

    def _load(self, i):
        data = self.loader.load(i)
        return OpenAPISpec.parse_obj(data)

    @property
    def _(self):
        return OperationIndex(self)

    def resolve_jr(self, root: "OpenAPISpec", obj, value: Reference):
        url, jp = JSONReference.split(value.ref)
        if url != "":
            url = pathlib.Path(url)
            if url not in self._cached:
                self._cached[url] = self._load(url)
            root = self._cached[url]

        try:
            return root.resolve_jp(jp)
        except ReferenceResolutionError as e:
            # add metadata to the error
            e.element = obj
            raise


class OpenAPISpec(ObjectExtended):
    """
    This class represents the root of the OpenAPI schema document, as defined
    in `the spec`_

    .. _the spec: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#openapi-object
    """

    openapi: str = Field(...)
    info: Info = Field(...)
    servers: Optional[List[Server]] = Field(default=None)
    paths: Dict[str, PathItem] = Field(required=True, default_factory=dict)
    components: Optional[Components] = Field(default_factory=Components)
    security: Optional[List[SecurityRequirement]] = Field(default=None)
    tags: Optional[List[Tag]] = Field(default=None)
    externalDocs: Optional[Dict[Any, Any]] = Field(default_factory=dict)

    def _resolve_references(self, api):
        """
        Resolves all reference objects below this object and notes their original
        value was a reference.
        """
        # don't circular import

        root = self

        def resolve(obj):
            if isinstance(obj, ObjectBase):
                for slot in filter(lambda x: not x.startswith("_"), obj.__fields_set__):
                    value = getattr(obj, slot)
                    if value is None:
                        continue

                    if isinstance(obj, PathItem) and slot == "ref":
                        ref = Reference.construct(ref=value)
                        ref._target = api.resolve_jr(root, obj, ref)
                        setattr(obj, slot, ref)

                    #                    if isinstance(obj, Discriminator) and slot == "mapping":
                    #                        mapping = dict()
                    #                        for k,v in value.items():
                    #                            mapping[k] = Reference.construct(ref=v)
                    #                        setattr(obj, slot, mapping)

                    value = getattr(obj, slot)
                    if isinstance(value, Reference):
                        value._target = api.resolve_jr(root, obj, value)
                    #                        setattr(obj, slot, resolved_value)
                    elif issubclass(type(value), ObjectBase):
                        # otherwise, continue resolving down the tree
                        resolve(value)
                    elif isinstance(value, dict):  # pydantic does not use Map
                        resolve(value)
                    elif isinstance(value, list):
                        # if it's a list, resolve its item's references
                        for item in value:
                            if isinstance(item, Reference):
                                item._target = api.resolve_jr(root, obj, item)
                            elif isinstance(item, (ObjectBase, dict, list)):
                                resolve(item)
                    elif isinstance(value, (str, int, float, datetime.datetime)):
                        continue
                    else:
                        raise TypeError(type(value))
            elif isinstance(obj, dict):
                for k, v in obj.items():
                    if isinstance(v, Reference):
                        if v.ref:
                            v._target = api.resolve_jr(root, obj, v)
                    elif isinstance(v, (ObjectBase, dict, list)):
                        resolve(v)

        resolve(self)

    def resolve_jp(self, jp):
        """
        Given a $ref path, follows the document tree and returns the given attribute.

        :param jp: The path down the spec tree to follow
        :type jp: str #/foo/bar

        :returns: The node requested
        :rtype: ObjectBase
        :raises ValueError: if the given path is not valid
        """
        path = jp.split("/")[1:]
        node = self

        for part in path:
            part = JSONPointer.decode(part)
            if isinstance(node, dict):
                if part not in node:  # pylint: disable=unsupported-membership-test
                    raise ReferenceResolutionError(f"Invalid path {path} in Reference")
                node = node.get(part)
            else:
                if not hasattr(node, part):
                    raise ReferenceResolutionError(f"Invalid path {path} in Reference")
                node = getattr(node, part)

        return node


class OperationIndex:
    class Iter:
        def __init__(self, spec):
            self.operations = []
            self.r = 0
            pi: PathItem
            for path, pi in spec.paths.items():
                op: Operation
                for method in pi.__fields_set__ & HTTP_METHODS:
                    op = getattr(pi, method)
                    if op.operationId is None:
                        continue
                    self.operations.append(op.operationId)
            self.r = iter(range(len(self.operations)))

        def __iter__(self):
            return self

        def __next__(self):
            return self.operations[next(self.r)]

    def __init__(self, api):
        self._api = api
        self._spec = api._spec

    def __getattr__(self, item):
        pi: PathItem
        for path, pi in self._spec.paths.items():
            op: Operation
            for method in pi.__fields_set__ & HTTP_METHODS:
                op = getattr(pi, method)
                if op.operationId != item:
                    continue

                if issubclass(self._api._session_factory, httpx.Client):
                    return Request(self._api, method, path, op)
                if issubclass(self._api._session_factory, httpx.AsyncClient):
                    return AsyncRequest(self._api, method, path, op)

        raise ValueError(item)

    def __iter__(self):
        return self.Iter(self._spec)


OpenAPISpec.update_forward_refs()



from typing import Union, Optional, Dict, Any

from pydantic import Field

from .example import Example
from .general import Reference
from .object_base import ObjectExtended
from .schemas import Schema


class Encoding(ObjectExtended):
    """
    A single encoding definition applied to a single schema property.

    .. _Encoding: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#encoding-object
    """

    contentType: Optional[str] = Field(default=None)
    headers: Optional[Dict[str, Union["Header", Reference]]] = Field(default_factory=dict)
    style: Optional[str] = Field(default=None)
    explode: Optional[bool] = Field(default=None)
    allowReserved: Optional[bool] = Field(default=None)


class MediaType(ObjectExtended):
    """
    A `MediaType`_ object provides schema and examples for the media type identified
    by its key.  These are used in a RequestBody object.

    .. _MediaType: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#media-type-object
    """

    schema_: Optional[Union[Schema, Reference]] = Field(required=True, alias="schema")
    example: Optional[Any] = Field(default=None)  # 'any' type
    examples: Optional[Dict[str, Union[Example, Reference]]] = Field(default_factory=dict)
    encoding: Optional[Dict[str, Encoding]] = Field(default_factory=dict)


from .parameter import Header

Encoding.update_forward_refs()



import re
from typing import Union, List, Optional, Dict

from pydantic import Field, BaseModel, root_validator

from .errors import SpecError
from .general import ExternalDocumentation
from .general import Reference
from .media import MediaType
from .object_base import ObjectBase, ObjectExtended
from .parameter import Header, Parameter
from .servers import Server


def _validate_parameters(op: "Operation", path):
    """
    Ensures that all parameters for this path are valid
    """
    assert isinstance(path, str)
    allowed_path_parameters = re.findall(r"{([a-zA-Z0-9\-\._~]+)}", path)

    for c in op.parameters:
        if c.in_ == "path":
            if c.name not in allowed_path_parameters:
                raise SpecError("Parameter name not found in path: {}".format(c.name))


class SecurityRequirement(BaseModel):
    """
    A `SecurityRequirement`_ object describes security schemes for API access.

    .. _SecurityRequirement: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#security-requirement-object
    """

    __root__: Dict[str, List[str]]

    @root_validator
    def validate_SecurityRequirement(cls, values):
        root = values.get("__root__", {})
        if not (len(root.keys()) == 1 and isinstance([c for c in root.values()][0], list) or len(root.keys()) == 0):
            raise ValueError(root)
        return values

    @property
    def name(self):
        if len(self.__root__.keys()):
            return list(self.__root__.keys())[0]
        return None

    @property
    def types(self):
        if self.name:
            return self.__root__[self.name]
        return None


class RequestBody(ObjectExtended):
    """
    A `RequestBody`_ object describes a single request body.

    .. _RequestBody: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#request-body-object
    """

    description: Optional[str] = Field(default=None)
    content: Dict[str, MediaType] = Field(...)
    required: Optional[bool] = Field(default=False)


class Link(ObjectExtended):
    """
    A `Link Object`_ describes a single Link from an API Operation Response to an API Operation Request

    .. _Link Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#link-object
    """

    operationRef: Optional[str] = Field(default=None)
    operationId: Optional[str] = Field(default=None)
    parameters: Optional[Dict[str, Union["RuntimeExpression", str]]] = Field(default=None)
    requestBody: Optional[dict] = Field(default=None)
    description: Optional[str] = Field(default=None)
    server: Optional[Server] = Field(default=None)

    @root_validator(pre=False)
    def validate_Link_operation(cls, values):
        if values["operationId"] != None and values["operationRef"] != None:
            raise SpecError("operationId and operationRef are mutually exclusive, only one of them is allowed")

        if values["operationId"] == values["operationRef"] == None:
            raise SpecError("operationId and operationRef are mutually exclusive, one of them must be specified")

        return values


class Response(ObjectExtended):
    """
    A `Response Object`_ describes a single response from an API Operation,
    including design-time, static links to operations based on the response.

    .. _Response Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#responses-object
    """

    description: str = Field(...)
    headers: Optional[Dict[str, Union[Header, Reference]]] = Field(default_factory=dict)
    content: Optional[Dict[str, MediaType]] = Field(default_factory=dict)
    links: Optional[Dict[str, Union[Link, Reference]]] = Field(default_factory=dict)


class Operation(ObjectExtended):
    """
    An Operation object as defined `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#operation-object
    """

    tags: Optional[List[str]] = Field(default=None)
    summary: Optional[str] = Field(default=None)
    description: Optional[str] = Field(default=None)
    externalDocs: Optional[ExternalDocumentation] = Field(default=None)
    operationId: Optional[str] = Field(default=None)
    parameters: List[Union[Parameter, Reference]] = Field(default_factory=list)
    requestBody: Optional[Union[RequestBody, Reference]] = Field(default=None)
    responses: Dict[str, Union[Response, Reference]] = Field(required=True)
    callbacks: Optional[Dict[str, Union["Callback", Reference]]] = Field(default_factory=dict)
    deprecated: Optional[bool] = Field(default=None)
    security: Optional[List[SecurityRequirement]] = Field(default_factory=list)
    servers: Optional[List[Server]] = Field(default=None)

    class Config:
        underscore_attrs_are_private = True


class PathItem(ObjectExtended):
    """
    A Path Item, as defined `here`_.
    Describes the operations available on a single path.

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#paths-object
    """

    ref: Optional[str] = Field(default=None, alias="$ref")
    summary: Optional[str] = Field(default=None)
    description: Optional[str] = Field(default=None)
    get: Optional[Operation] = Field(default=None)
    put: Optional[Operation] = Field(default=None)
    post: Optional[Operation] = Field(default=None)
    delete: Optional[Operation] = Field(default=None)
    options: Optional[Operation] = Field(default=None)
    head: Optional[Operation] = Field(default=None)
    patch: Optional[Operation] = Field(default=None)
    trace: Optional[Operation] = Field(default=None)
    servers: Optional[List[Server]] = Field(default=None)
    parameters: Optional[List[Union[Parameter, Reference]]] = Field(default_factory=list)


class Callback(ObjectBase):
    """
    A map of possible out-of band callbacks related to the parent operation.

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#callback-object

    This object MAY be extended with Specification Extensions.
    """

    __root__: Dict[str, PathItem]


class RuntimeExpression(ObjectBase):
    """


    .. Runtime Expression: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#runtime-expressions
    """

    __root__: str = Field(...)


Operation.update_forward_refs()
Link.update_forward_refs()



from typing import Optional

from pydantic import Field

from .object_base import ObjectExtended


class Contact(ObjectExtended):
    """
    Contact object belonging to an Info object, as described `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#contact-object
    """

    email: str = Field(default=None)
    name: str = Field(default=None)
    url: str = Field(default=None)


class License(ObjectExtended):
    """
    License object belonging to an Info object, as described `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#license-object
    """

    name: str = Field(...)
    url: Optional[str] = Field(default=None)


class Info(ObjectExtended):
    """
    An OpenAPI Info object, as defined in `the spec`_.

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#info-object
    """

    title: str = Field(...)
    description: Optional[str] = Field(default=None)
    termsOfService: Optional[str] = Field(default=None)
    license: Optional[License] = Field(default=None)
    contact: Optional[Contact] = Field(default=None)
    version: str = Field(...)



import types
import uuid
from typing import Union, List, Any, Optional, Dict, Literal, Annotated

from pydantic import Field, root_validator, Extra, BaseModel

from .general import Reference  # need this for Model below
from .object_base import ObjectExtended
from .xml import XML

TYPE_LOOKUP = {
    "array": list,
    "integer": int,
    "object": dict,
    "string": str,
    "boolean": bool,
}


class Discriminator(ObjectExtended):
    """

    .. here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#discriminator-object
    """

    propertyName: str = Field(...)
    mapping: Optional[Dict[str, str]] = Field(default_factory=dict)


class Schema(ObjectExtended):
    """
    The `Schema Object`_ allows the definition of input and output data types.

    .. _Schema Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#schema-object
    """

    title: Optional[str] = Field(default=None)
    multipleOf: Optional[int] = Field(default=None)
    maximum: Optional[float] = Field(default=None)  # FIXME Field(discriminator='type') would be better
    exclusiveMaximum: Optional[bool] = Field(default=None)
    minimum: Optional[float] = Field(default=None)
    exclusiveMinimum: Optional[bool] = Field(default=None)
    maxLength: Optional[int] = Field(default=None)
    minLength: Optional[int] = Field(default=None)
    pattern: Optional[str] = Field(default=None)
    maxItems: Optional[int] = Field(default=None)
    minItems: Optional[int] = Field(default=None)
    uniqueItems: Optional[bool] = Field(default=None)
    maxProperties: Optional[int] = Field(default=None)
    minProperties: Optional[int] = Field(default=None)
    required: Optional[List[str]] = Field(default_factory=list)
    enum: Optional[list] = Field(default=None)

    type: Optional[str] = Field(default=None)
    allOf: Optional[List[Union["Schema", Reference]]] = Field(default_factory=list)
    oneOf: Optional[List[Union["Schema", Reference]]] = Field(default_factory=list)
    anyOf: Optional[List[Union["Schema", Reference]]] = Field(default_factory=list)
    not_: Optional[Union["Schema", Reference]] = Field(default=None, alias="not")
    items: Optional[Union["Schema", Reference]] = Field(default=None)
    properties: Optional[Dict[str, Union["Schema", Reference]]] = Field(default_factory=dict)
    additionalProperties: Optional[Union[bool, "Schema", Reference]] = Field(default=None)
    description: Optional[str] = Field(default=None)
    format: Optional[str] = Field(default=None)
    default: Optional[str] = Field(default=None)  # TODO - str as a default?
    nullable: Optional[bool] = Field(default=None)
    discriminator: Optional[Discriminator] = Field(default=None)  # 'Discriminator'
    readOnly: Optional[bool] = Field(default=None)
    writeOnly: Optional[bool] = Field(default=None)
    xml: Optional[XML] = Field(default=None)  # 'XML'
    externalDocs: Optional[dict] = Field(default=None)  # 'ExternalDocs'
    example: Optional[Any] = Field(default=None)
    deprecated: Optional[bool] = Field(default=None)
    #    contentEncoding: Optional[str] = Field(default=None)
    #    contentMediaType: Optional[str] = Field(default=None)
    #    contentSchema: Optional[str] = Field(default=None)

    _model_type: object
    _request_model_type: object

    """
    The _identity attribute is set during OpenAPI.__init__ and used at get_type()
    """
    _identity: str

    class Config:
        #        keep_untouched = (lru_cache,)
        extra = Extra.forbid

    @root_validator
    def validate_Schema_number_type(cls, values: Dict[str, object]):
        conv = ["minimum", "maximum"]
        if values.get("type", None) == "integer":
            for i in conv:
                v = values.get(i, None)
                if v is not None:
                    values[i] = int(v)
        return values

    #    @lru_cache
    def get_type(self, names: List[str] = None, discriminators: List[Discriminator] = None):
        return Model.from_schema(self, names, discriminators)

    def model(self, data: Dict):
        """
        Generates a model representing this schema from the given data.

        :param data: The data to create the model from.  Should match this schema.
        :type data: dict

        :returns: A new :any:`Model` created in this Schema's type from the data.
        :rtype: self.get_type()
        """
        if self.type in ("string", "number"):
            assert len(self.properties) == 0
            # more simple types
            # if this schema represents a simple type, simply return the data
            # TODO - perhaps assert that the type of data matches the type we
            # expected
            return data
        elif self.type == "array":
            return [self.items.get_type().parse_obj(i) for i in data]
        else:
            return self.get_type().parse_obj(data)


class Model(BaseModel):
    class Config:
        extra: Extra.forbid

    @classmethod
    def from_schema(cls, shma: Schema, shmanm: List[str] = None, discriminators: List[Discriminator] = None):

        if shmanm is None:
            shmanm = []

        if discriminators is None:
            discriminators = []

        def typeof(schema: Schema):
            r = None
            if schema.type == "integer":
                r = int
            elif schema.type == "number":
                r = float
            elif schema.type == "string":
                r = str
            elif schema.type == "boolean":
                r = bool
            elif schema.type == "array":
                r = List[schema.items.get_type()]
            elif schema.type == "object":
                return schema.get_type()
            elif schema.type is None:  # discriminated root
                return None
            else:
                raise TypeError(schema.type)

            return r

        def annotationsof(schema: Schema):
            annos = dict()
            if schema.type == "array":
                annos["__root__"] = typeof(schema)
            else:

                for name, f in schema.properties.items():
                    r = None
                    for discriminator in discriminators:
                        if name != discriminator.propertyName:
                            continue
                        for disc, v in discriminator.mapping.items():
                            if v in shmanm:
                                r = Literal[disc]
                                break
                        else:
                            raise ValueError(schema)
                        break
                    else:
                        r = typeof(f)
                    if name not in schema.required:
                        annos[name] = Optional[r]
                    else:
                        annos[name] = r
            return annos

        def fieldof(schema: Schema):
            r = dict()
            if schema.type == "array":
                return r
            else:
                for name, f in schema.properties.items():
                    args = dict()
                    for i in ["enum", "default"]:
                        v = getattr(f, i, None)
                        if v:
                            args[i] = v
                    r[name] = Field(**args)
            return r

        # do not create models for primitive types
        if shma.type in ("string", "integer"):
            return typeof(shma)

        type_name = shma.title or shma._identity if hasattr(shma, "_identity") else str(uuid.uuid4())
        namespace = dict()
        annos = dict()
        if shma.allOf:
            for i in shma.allOf:
                annos.update(annotationsof(i))
        elif shma.anyOf:
            t = tuple(
                [
                    i.get_type(names=shmanm + [i.ref], discriminators=discriminators + [shma.discriminator])
                    for i in shma.anyOf
                ]
            )
            if shma.discriminator and shma.discriminator.mapping:
                annos["__root__"] = Annotated[Union[t], Field(discriminator=shma.discriminator.propertyName)]
            else:
                annos["__root__"] = Union[t]
        elif shma.oneOf:
            t = tuple(
                [
                    i.get_type(names=shmanm + [i.ref], discriminators=discriminators + [shma.discriminator])
                    for i in shma.oneOf
                ]
            )
            if shma.discriminator and shma.discriminator.mapping:
                annos["__root__"] = Annotated[Union[t], Field(discriminator=shma.discriminator.propertyName)]
            else:
                annos["__root__"] = Union[t]
        else:
            annos = annotationsof(shma)
            namespace.update(fieldof(shma))

        namespace["__annotations__"] = annos

        m = types.new_class(type_name, (BaseModel,), {}, lambda ns: ns.update(namespace))
        m.update_forward_refs()
        return m


Schema.update_forward_refs()



from typing import List, Optional, Dict

from pydantic import Field

from .object_base import ObjectExtended


class ServerVariable(ObjectExtended):
    """
    A ServerVariable object as defined `here`_.

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#server-variable-object
    """

    enum: Optional[List[str]] = Field(default=None)
    default: str = Field(...)
    description: Optional[str] = Field(default=None)


class Server(ObjectExtended):
    """
    The Server object, as described `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#server-object
    """

    url: str = Field(...)
    description: Optional[str] = Field(default=None)
    variables: Optional[Dict[str, ServerVariable]] = Field(default_factory=dict)



from typing import Optional

from pydantic import BaseModel, Field, root_validator, Extra


class ObjectBase(BaseModel):
    """
    The base class for all schema objects.  Includes helpers for common schema-
    related functions.
    """

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = False
        extra = Extra.forbid


class ObjectExtended(ObjectBase):
    extensions: Optional[object] = Field(default=None)

    @root_validator(pre=True)
    def validate_ObjectExtended_extensions(cls, values):
        """FIXME
        https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#specification-extensions
        :param values:
        :return: values
        """
        e = dict()
        for k, v in values.items():
            if k.startswith("x-"):
                e[k[2:]] = v
        if len(e):
            for i in e.keys():
                del values[f"x-{i}"]
            if "extensions" in values.keys():
                raise ValueError("extensions")
            values["extensions"] = e

        return values



import sys
import yaml
from pathlib import Path

from .openapi import OpenAPI

from .loader import FileSystemLoader


def main():
    name = sys.argv[1]
    loader = FileSystemLoader(Path().cwd())
    spec = loader.load(name)
    try:
        OpenAPI(name, spec, loader=loader)
    except ValueError as e:
        print(e)
    else:
        print("OK")


if __name__ == "__main__":
    main()



from typing import Optional

from pydantic import Field

from .object_base import ObjectExtended
from .general import ExternalDocumentation


class Tag(ObjectExtended):
    """
    A `Tag Object`_ holds a reusable set of different aspects of the OAS
    spec.

    .. _Tag Object: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#tag-object
    """

    name: str = Field(...)
    description: Optional[str] = Field(default=None)
    externalDocs: Optional[ExternalDocumentation] = Field(default=None)
